#!/usr/bin/env python

# Parts of this code are
# Copyright (C) 2002  Juan David Ibanez Palomar <j-david@noos.fr>

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# Extensions added by Dave Kuhlman <dkuhlman@rexx.com>.

"""
Flux lets to describe graphs and objects that move from one to
another state of the graph following the defined transitions.

Flux provides an interface for applications to associate specific
semantics to states and transitions.

It can be used for Workflow applications, and maybe for other
purposes.
"""

import sys, getopt
from xml.dom import minidom
from xml.dom import Node

__version__ = "$Revision: 1.5 $"

FluxError='FluxError'


class Workflow:
    """
    This class is used to describe a workflow (actually it's just a
    graph). A workflow has states (one of them is the initial state),
    and states have transitions that go to another state.
    """

    def __init__(self):
        """
        Initialize the workflow.
        """
        self.name = ''
        self.states = {}
        self.state_defaultinit = None

    def getName(self): return self.name
    def getStates(self): return self.states
    def getState_defaultinit(self): return self.state_defaultinit

    def fx_addstate(self, ident):
        """
        Adds a new state.
        """
        state = State()
        state.initialize(ident)
        self.states[ident] = state


    def fx_initstate(self, id):
        """
        Sets the default initial state.
        """
        if not self.states.has_key(id):
            raise FluxError, "invalid initial state: '%s'" % id
        self.state_defaultinit = id


    def fx_addtrans(self, ident, state_from_id, state_to_id, condition=None, action=None):
        """
        Adds a new transition, 'state_from' and 'state_to' are
        respectively the origin and destination states of the
        transition.
        """
        try:    
            state_from = self.states[state_from_id]
        except KeyError:
            raise FluxError, "unregistered state: '%s'" % state_from
        try:
            state_to = self.states[state_to_id]
        except KeyError:
            raise FluxError, "unregistered state: '%s'" % state_to
        transition = Transition()
        transition.initialize(ident, state_from_id, state_to_id)
        transition.setCondition(condition)
        transition.setAction(action)
        state_from.fx_addtrans(ident, transition)

    def export(self, outfile, level):
        showIndent(outfile, level)
        outfile.write('<workflow>\n')
        level += 1
        showIndent(outfile, level)
        outfile.write('<name>%s</name>\n' % self.name)
        showIndent(outfile, level)
        outfile.write('<state_initial>%s</state_initial>\n' % self.state_defaultinit)
        for ident, state in self.states.iteritems():
            state.export(outfile, level)
        level -= 1
        showIndent(outfile, level)
        outfile.write('</workflow>\n')

    def build(self, node_):
        attrs = node_.attributes
        for child in node_.childNodes:
            if child.nodeType == Node.ELEMENT_NODE and \
                child.nodeName == 'state_initial':
                state_defaultinit = ''
                for text_ in child.childNodes:
                    state_defaultinit += text_.nodeValue
                self.state_defaultinit = state_defaultinit
            elif child.nodeType == Node.ELEMENT_NODE and \
                child.nodeName == 'name':
                name = ''
                for text_ in child.childNodes:
                    name += text_.nodeValue
                self.name = name
            elif child.nodeType == Node.ELEMENT_NODE and \
                child.nodeName == 'state':
                obj = State()
                ident = obj.build(child)
                self.states[ident] = obj

    def genModuleClass(self, outFile):
        for state in self.getStates().itervalues():
            state.genModuleClass(outFile)

    def genModuleFunction(self, outFile):
        for state in self.getStates().itervalues():
            state.genModuleFunction(outFile)

    def genStateDict(self, outFile):
        s1 = "#\n"
        outFile.write(s1)
        s1 = "# Dictionary to map state names to state implementations\n"
        outFile.write(s1)
        s1 = "#\n"
        outFile.write(s1)
        s1 = "StateDict = {\n"
        outFile.write(s1)
        for state in self.getStates().itervalues():
            name = state.getId()
            s1 = "    '%s': %s,\n" % (name, name)
            outFile.write(s1)
        s1 = "    }\n\n"
        outFile.write(s1)

# end class Workflow


class State:
    """
    This class is used to describe a state. An state has transitions
    to other states.
    """

    def __init__(self):
        """
        Initialize the state.
        """
        self.id = ''
        self.transitions = {}

    def initialize(self, ident):
        self.id = ident

    def getId(self): return self.id
    def getTransitions(self): return self.transitions

    def fx_addtrans(self, ident, transition):
        """
        Adds a new transition.
        """
        self.transitions[ident] = transition

    def export(self, outfile, level):
        showIndent(outfile, level)
        outfile.write('<state>\n')
        level += 1
        showIndent(outfile, level)
        outfile.write('<id>%s</id>\n' % self.id)
        for transition in self.transitions.itervalues():
            transition.export(outfile, level)
        level -= 1
        showIndent(outfile, level)
        outfile.write('</state>\n')

    def build(self, node_):
        ident = ''
        attrs = node_.attributes
        for child in node_.childNodes:
            if child.nodeType == Node.ELEMENT_NODE and \
                child.nodeName == 'id':
                ident = ''
                for text_ in child.childNodes:
                    ident += text_.nodeValue
                self.id = ident
            elif child.nodeType == Node.ELEMENT_NODE and \
                child.nodeName == 'pagename':
                pagename = ''
                for text_ in child.childNodes:
                    pagename += text_.nodeValue
                self.pagename = pagename
            elif child.nodeType == Node.ELEMENT_NODE and \
                child.nodeName == 'transition':
                obj = Transition()
                transIdent = obj.build(child)
                self.transitions[transIdent] = obj
        return ident

    def genModuleClass(self, outFile):
        s1 = "class %s(State):\n" % self.getId()
        outFile.write(s1)
        s1 = "    def __init__(self, doc):\n"
        outFile.write(s1)
        s1 = "        State.__init__(self, doc)\n"
        outFile.write(s1)
        s1 = "    def _q_index(self, request):\n"
        outFile.write(s1)
        s1 = "        nextState = None\n"
        outFile.write(s1)
        for transition in self.getTransitions().itervalues():
            transition.genModuleClass(outFile)
        s1 = "        self.doc.setFsmcurrentstate(nextState)\n"
        outFile.write(s1)
        s1= "        content = self.generateContent()\n"
        outFile.write(s1)
        s1= "        self.setXmlResponse(request)\n"
        outFile.write(s1)
        s1= "        return content\n"
        outFile.write(s1)
        s1 = "\n\n"
        outFile.write(s1)

    def genModuleFunction(self, outFile):
        s1 = "def %s(request, doc):\n" % self.getId()
        outFile.write(s1)
        s1 = "    nextState = None\n"
        outFile.write(s1)
        for transition in self.getTransitions().itervalues():
            transition.genModuleFunction(outFile)
        s1 = "    doc.setFsmcurrentstate(doc, nextState)\n"
        outFile.write(s1)
        s1= "    content = generateContent(doc)\n"
        outFile.write(s1)
        s1= "    setXmlResponse(request)\n"
        outFile.write(s1)
        s1= "    return content\n"
        outFile.write(s1)
        s1 = "\n\n"
        outFile.write(s1)

# end class State


class Transition:
    """
    This class is used to describe transitions. Transitions come from
    one state and go to another.
    """
    def __init__(self):
        self.id = ''
        self.state_from = ''
        self.state_to = ''
        self.condition = None
        self.action = None

    def getId(self): return self.id
    def getState_from_id(self): return self.state_from
    def getState_to_id(self): return self.state_to
    def getCondition(self): return self.condition
    def setCondition(self, condition): self.condition = condition
    def getAction(self): return self.action
    def setAction(self, action): self.action = action

    def initialize(self, ident, state_from, state_to):
        """
        Initialize the transition.
        """
        self.id = ident
        self.state_from = state_from
        self.state_to = state_to

    def export(self, outfile, level):
        showIndent(outfile, level)
        outfile.write('<transition>\n')
        level += 1
        showIndent(outfile, level)
        outfile.write('<id>%s</id>\n' % self.id)
        showIndent(outfile, level)
        outfile.write('<state_from>%s</state_from>\n' % self.state_from)
        showIndent(outfile, level)
        outfile.write('<state_to>%s</state_to>\n' % self.state_to)
        showIndent(outfile, level)
        if self.condition:
            outfile.write('<condition>%s</condition>\n' % self.condition)
        else:
            outfile.write('<condition></condition>\n')
        showIndent(outfile, level)
        if self.action:
            outfile.write('<action>%s</action>\n' % self.action)
        else:
            outfile.write('<action></action>\n')
        level -= 1
        showIndent(outfile, level)
        outfile.write('</transition>\n')

    def build(self, node_):
        attrs = node_.attributes
        ident = ''
        for child in node_.childNodes:
            if child.nodeType == Node.ELEMENT_NODE and \
                child.nodeName == 'id':
                ident = ''
                for text_ in child.childNodes:
                    ident += text_.nodeValue
                self.id = ident
            elif child.nodeType == Node.ELEMENT_NODE and \
                child.nodeName == 'state_from':
                state_from = ''
                for text_ in child.childNodes:
                    state_from += text_.nodeValue
                self.state_from = state_from
            elif child.nodeType == Node.ELEMENT_NODE and \
                child.nodeName == 'state_to':
                state_to = ''
                for text_ in child.childNodes:
                    state_to += text_.nodeValue
                self.state_to = state_to
            elif child.nodeType == Node.ELEMENT_NODE and \
                child.nodeName == 'condition':
                condition = ''
                for text_ in child.childNodes:
                    condition += text_.nodeValue
                self.condition = condition
            elif child.nodeType == Node.ELEMENT_NODE and \
                child.nodeName == 'action':
                action = ''
                for text_ in child.childNodes:
                    action += text_.nodeValue
                self.action = action
        return ident

    def genModuleClass(self, outFile):
        s1 = "        # Transition %s\n" % self.id
        outFile.write(s1)
        if self.condition:
            s1 = "        if %s:\n" % self.condition
            outFile.write(s1)
            s1 = "            nextState = '%s'\n" % self.state_to
            outFile.write(s1)
            if self.action:
                s1 = "            %s\n" % self.action
                outFile.write(s1)
        else:
            s1 = "        nextState = '%s'\n" % self.state_to
            outFile.write(s1)
            if self.action:
                s1 = "        %s\n" % self.action
                outFile.write(s1)

    def genModuleFunction(self, outFile):
        s1 = "    # Transition %s\n" % self.id
        outFile.write(s1)
        if self.condition:
            s1 = "    if %s:\n" % self.condition
            outFile.write(s1)
            s1 = "        nextState = '%s'\n" % self.state_to
            outFile.write(s1)
            if self.action:
                s1 = "        %s\n" % self.action
                outFile.write(s1)
        else:
            s1 = "    nextState = '%s'\n" % self.state_to
            outFile.write(s1)
            if self.action:
                s1 = "    %s\n" % self.action
                outFile.write(s1)

# end class Transition


class WorkflowAware:
    """
    Mixin class to be used for workflow aware objects. The instances of
    a class that inherits from WorkflowAware can be "within" the workflow,
    this means that they keep track of the current state of the object.

    Specific application semantics for states and transitions can be
    implemented as methods of the WorkflowAware-derived "developer 
    class". These methods get associated with the individual
    states and transitions by a simple naming scheme. For example, 
    if a workflow has two states 'private' and 'public', and a 
    transition 'publish' that goes from 'private' to 'public', 
    the following happens when the transition is executed:
    1. if implemented, the method 'fx_onleave_private' is called
       (it is called each time the object leaves the 'private' state)
    2. if implemented, the method 'fx_ontrans_publish' is called
       (it is called whenever this transition is executed)
    3. if implemented, the method 'fx_onenter_public' is called
       (it is called each time the object enters the 'public' state)
    These state/transition "handlers" can also be passed arguments
    from the caller of the transition; for instance, in a web-based
    system it might be useful to have the HTTP request that triggered 
    the current transition available in the handlers.
    
    """

    def __init__(self, workflow, initstate=None, *args, **kw):
        """
        Initialize the workflow aware object. initstate is 
        the workflow state that should be taken on initially 
        (if omitted or None, the workflow must provide a default 
        initial state). Extra arguments are passed to the 
        enter-state handler (if any) of the initial state. 
        """
        # extra constructor arguments are passed trough to enter-state handler
        # for initial state, if any
        self.fx_enterworkflow(workflow, initstate, *args, **kw)


    def fx_enterworkflow(self, workflow, initstate=None, *args, **kw):
        """
        [Re-]Bind this object to a specific workflow. initstate is 
        the workflow state that should be taken on initially 
        (if omitted or None, the workflow must provide a default 
        initial state). Extra arguments args are passed to the 
        enter-state handler (if any) of the initial state. 
        """
        # extra arguments are passed trough to enter-state handler
        # for initial state, if any
        self.flux_workflow = workflow
        if initstate is None: initstate = workflow.state_defaultinit
        if not initstate:
            raise FluxError, 'undefined initial workflow state'
        if not workflow.states.has_key(initstate):
            raise FluxError, "invalid initial state: '%s'" % initstate
        self.flux_state=initstate
        # call app-specific enter-state handler for initial state, if any
        name = 'fx_onenter_%s' % initstate
        if hasattr(self, name):
            getattr(self, name)(*args, **kw)


    def fx_dotrans(self, transid, *args, **kw):
        """
        Performs a transition, changes the state of the object and
        runs any defined state/transition handlers. Extra 
        arguments are passed down to all handlers called.
        """
        # Get the workflow
        workflow = self.flux_workflow

        # Get the current state
        state = workflow.states[self.flux_state]
        
        try:
            # Get the new state id
            state = state.transitions[transid].state_to
        except KeyError:
            raise FluxError, "transition '%s' is invalid from state '%s'" % (transid, self.flux_state)
        
        # call app-specific leave- state  handler if any
        name = 'fx_onleave_%s' % self.flux_state
        if hasattr(self, name):
            getattr(self, name)(*args, **kw)

        # Set the new state
        self.flux_state = state

        # call app-specific transition handler if any
        name = 'fx_ontrans_%s' % transid
        if hasattr(self, name):
            getattr(self, name)(*args, **kw)

        # call app-specific enter-state handler if any
        name = 'fx_onenter_%s' % state
        if hasattr(self, name):
            getattr(self, name)(*args, **kw)
            
    def fx_getstate(self):
        """return id of current state"""
        return self.flux_state
# end class WorkflowAware



class Token(WorkflowAware):
    """
    This class should be used when the document can be in different
    states at the same time (likely states that belong to different
    workflows).

    In this situation the class shouldn't inherit from WorkflowAware,
    instead it should contain as many tokens as needed, which would
    be the ones that follow the workflow.
    """
# end class Token


#
# Utility functions.
#

def showIndent(outfile, level):
    for idx in range(level):
        outfile.write('    ')


#
# Test harness.
#
def test1():
    wf = Workflow()
    wf.fx_addstate('Start')
    wf.fx_addstate('Finish')
    wf.fx_addstate('Plant')
    wf.fx_addstate('Fruit')
    wf.fx_addstate('Vegetable')
    wf.fx_addstate('Continue')
    wf.fx_initstate('Start')
    wf.fx_addtrans('t1', 'Start', 'Plant', action='logon()')
    wf.fx_addtrans('t2', 'Plant', 'Fruit', condition='self.is_it("fruit")')
    wf.fx_addtrans('t3', 'Plant', 'Vegetable', condition='self.is_it("vegetable")')
    wf.fx_addtrans('t4', 'Fruit', 'Continue', action='self.show_fruit()')
    wf.fx_addtrans('t5', 'Vegetable', 'Continue', action='self.show_vegetable()')
    #wf.fx_addtrans('t6', 'Vegetable', 'Continue')
    wf.fx_addtrans('t7', 'Continue', 'Finish',
        condition='is_it("quit")', action='logoff()')
    wf.fx_addtrans('t8', 'Continue', 'Plant', condition='is_it("repeat")')
    sys.stdout.write('<?xml version="1.0" ?>\n')
    wf.export(sys.stdout, 0)


def test2(inFileName):
    doc = minidom.parse(inFileName)
    rootNode = doc.childNodes[0]
    wf = Workflow()
    wf.build(rootNode)
    sys.stdout.write('<?xml version="1.0" ?>\n')
    wf.export(sys.stdout, 0)


HEADER_module_class = """

# Import the parser and data structures for the XML interchange document.
import %s

_q_exports = []


# StateDict must be defined below after states are defined.
def getStateDict():
    return StateDict


def _q_index(request):
    buf = request.stdin
    content = buf.read()
    doc = statesub.parseString(content)
    state = doc.getFsmcurrentstate()
    arg1 = doc.getFsminput().getArg1()
    cls = getStateDict().get(state, None)
    if cls:
        obj = cls(doc)
        return obj._q_index(request)
    else:
        raise TraversalError('No such state: %%s' %% state)


#
# State implementations.
#

class State:
    def __init__(self, doc):
        self.doc = doc
    def setXmlResponse(self, request):
        response = request.response
        response.set_header('Content-type', 'text/xml; charset=iso-8859-1')
    def generateContent(self):
        contentStream = StringIO.StringIO()
        self.doc.export(contentStream, 0)
        content = contentStream.getvalue()
        contentStream.close()
        return content

"""

HEADER_module_function = """

# Import the parser and data structures for the XML interchange document.
import %s

_q_exports = []


# StateDict must be defined below after states are defined.
def getStateDict():
    return StateDict


def _q_index(request):
    buf = request.stdin
    content = buf.read()
    doc = statesub.parseString(content)
    state = doc.getFsmcurrentstate()
    arg1 = doc.getFsminput().getArg1()
    cls = getStateDict().get(state, None)
    if cls:
        obj = cls(doc)
        return obj._q_index(request)
    else:
        raise TraversalError('No such state: %%s' %% state)


#
# State implementations.
#

"""

FOOTER_module_class = """
"""

FOOTER_module_function = """
#
# Utility functions.
#

def generateContent(doc):
    contentStream = StringIO.StringIO()
    doc.export(contentStream, 0)
    content = contentStream.getvalue()
    contentStream.close()
    return content

def setXmlResponse(request):
    response = request.response
    response.set_header('Content-type', 'text/xml; charset=iso-8859-1')

"""

def genModuleClass(inFileName, outName, stateModule):
    fsm = parse(inFileName)
    outFile = file(outName, 'w')
    outFile.write(HEADER_module_class % stateModule)
    fsm.genModuleClass(outFile)
    fsm.genStateDict(outFile)
    outFile.write(FOOTER_module_class)
    outFile.close()


def genModuleFunction(inFileName, outName, stateModule):
    fsm = parse(inFileName)
    outFile = file(outName, 'w')
    outFile.write(HEADER_module_function % stateModule)
    fsm.genModuleFunction(outFile)
    fsm.genStateDict(outFile)
    outFile.write(FOOTER_module_function)
    outFile.close()


def parse(inFilename):
    doc = minidom.parse(inFilename)
    rootNode = doc.childNodes[0]
    wf = Workflow()
    wf.build(rootNode)
    return wf


def parseString(inString):
    doc = minidom.parseString(inString)
    rootNode = doc.childNodes[0]
    wf = Workflow()
    wf.build(rootNode)
    return wf


USAGE_TEXT = """
Usage: python fluxlib.py [ options ] <infilename> [ <outname> ]

Options:
    -s, --state <module>    Module implementing parser and data structures
                                for XML doc.
    -m, --mode <mode>       Modes -- See below.

Modes:
    module_class            Generate Quixote/FSM/REST app.
                                Container: module  State: class
    module_function         Generate Quixote/FSM/REST app.
                                Container: module  State: function
    test1                   Generate some classes and XML output.
    test2                   Read an XML file, create classes, and re-generate
                                XML output.

Examples:
    python fsmlib.py -m test1
    python fsmlib.py -m test1 > workflow1.xml
    python fsmlib.py -m test2 workflow1.xml
    python fsmlib.py -m module_class workflow1.xml workflow1.py
    python fsmlib.py -m module_function workflow1.xml workflow1.py
"""

def usage():
    print USAGE_TEXT
    sys.exit(-1)


def export():
    args = sys.argv[1:]
    if len(args) != 1:
        usage()
    infilename = args[0]
    root = parse(infilename)
    sys.stdout.write('?xml version="1.0" ?>\n')
    root.export(sys.stdout, 0)


def main():
    args = sys.argv[1:]
    try:
        opts, args = getopt.getopt(args, 'hm:s:', ['help', 'mode=', 'state='])
    except:
        usage()
    relink = 1
    mode = None
    stateModule = 'statesub'
    for opt, val in opts:
        if opt in ('-h', '--help'):
            usage()
        elif opt in ('-m', '--mode'):
            mode = val
        elif opt in ('-s', '--state'):
            stateModule = val
    if mode == 'test1':
        if len(args) != 0:
            usage()
        test1()
    elif mode == 'test2':
        if len(args) != 1:
            usage()
        test2(args[0])
    elif mode == 'module_class':
        if len(args) != 2:
            usage()
        inFileName = args[0]
        outName = args[1]
        genModuleClass(inFileName, outName, stateModule)
    elif mode == 'module_function':
        if len(args) != 2:
            usage()
        inFileName = args[0]
        outName = args[1]
        genModuleFunction(inFileName, outName, stateModule)
    else:
        usage()


if __name__ == '__main__':
    main()
    #import pdb
    #pdb.run('main()')


